from typing import Tuple, Callable, Dict, List
from functools import wraps
import codecs
from .strict_type import *

T = TypeVar("T")
global_encoding: str = "utf-8"


def _lambda_raise_(ex):
    raise ex


def check_encoding(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        for key, value in kwargs.items():
            if key == "encoding":
                try:
                    codecs.lookup(value)
                except LookupError:
                    raise Exception(f"{value} is not a proper encoding.")
                return func(*args, **kwargs)
        return func(*args, **kwargs)

    return func_wrapper


pack_dict: Dict[type, Callable[[Any], bytes]] = {
    int: (lambda obj: load(Int(obj))),
    bool: (lambda obj: struct.pack("=?", obj)),
    float: (lambda obj: load(Float(obj))),
    Char: (lambda obj: obj.__pack__()),
    UnsignedChar: (lambda obj: obj.__pack__()),
    SignedChar: (lambda obj: obj.__pack__()),
    Short: (lambda obj: obj.__pack__()),
    UnsignedShort: (lambda obj: obj.__pack__()),
    Int: (lambda obj: obj.__pack__()),
    UnsignedInt: (lambda obj: obj.__pack__()),
    Long: (lambda obj: obj.__pack__()),
    UnsignedLong: (lambda obj: obj.__pack__()),
    LongLong: (lambda obj: obj.__pack__()),
    UnsignedLongLong: (lambda obj: obj.__pack__()),
    Float: (lambda obj: obj.__pack__()),
    Double: (lambda obj: obj.__pack__()),
    bytes: (lambda obj: struct.pack("i{}s".format(len(obj)), len(obj), obj)),
    str: (lambda obj: load(obj.encode(global_encoding))),
    list: (lambda obj: _pack_list(obj)),
    set: (lambda obj: _lambda_raise_(NotImplementedError)),
    dict: (lambda obj: _lambda_raise_(NotImplementedError)),
    tuple: (lambda obj: load(*obj)),
}

unpack_dict: Dict[type, Callable[[T, bytes], T]] = {
    int: (lambda obj, data: _unpack_helper(data, Int(obj))),
    bool: (lambda _, data: (struct.unpack("=?", data[:1])[0], data[1:],)),
    float: (lambda obj, data: _unpack_helper(data, Float(obj))),
    Char: (lambda obj, data: obj.__unpack__(data)),
    UnsignedChar: (lambda obj, data: obj.__unpack__(data)),
    SignedChar: (lambda obj, data: obj.__unpack__(data)),
    Short: (lambda obj, data: obj.__unpack__(data)),
    UnsignedShort: (lambda obj, data: obj.__unpack__(data)),
    Int: (lambda obj, data: obj.__unpack__(data)),
    UnsignedInt: (lambda obj, data: obj.__unpack__(data)),
    Long: (lambda obj, data: obj.__unpack__(data)),
    UnsignedLong: (lambda obj, data: obj.__unpack__(data)),
    LongLong: (lambda obj, data: obj.__unpack__(data)),
    UnsignedLongLong: (lambda obj, data: obj.__unpack__(data)),
    Float: (lambda obj, data: obj.__unpack__(data)),
    Double: (lambda obj, data: obj.__unpack__(data)),
    bytes: (lambda _, data: _unpack_bytes(data)),
    str: (lambda _, data: _unpack_string(data)),
    list: (lambda obj, data: _unpack_list(data, obj)),
    set: (lambda obj, _: _lambda_raise_(NotImplementedError)),
    dict: (lambda obj, _: _lambda_raise_(NotImplementedError)),
    tuple: (lambda obj, data: _unpack_tuple(data, obj)),
}


def _pack_list(li: List[T]) -> bytes:
    """
    Packs *li* of size *n* in the following format:
    ``[n][li_1][li_2]...[li_n]``

    :param li: List to be packed
    :return: Packed byte string of *list*
    """
    if len(li) == 0:
        return load(len(li))
    conformed_list = _conform_list(li)
    return b"".join([load(i) for i in [len(conformed_list)] + conformed_list])


def _conform_list(li: List[Any]) -> List[T]:
    """
    Ensures that every element in *li* can conform to one type
    :param li: list to conform
    :return: conformed list
    """
    conform_type = li[0].__class__
    for i in li:
        if isinstance(i, StrictType):
            conform_type = i.__class__
            break
    base_type = (
        conform_type.__base__ if conform_type.__base__ != object else None
    )  # do not let base_type be 'object'
    if not all(type(i) == conform_type or type(i) == base_type for i in li):
        raise Exception(f"{li} can not be conformed to the {conform_type}")
    return [i if isinstance(i, conform_type) else conform_type(i) for i in li]


def _unpack_string(data: bytes) -> (str, bytes):
    result, data = _unpack_bytes(data)
    return result.decode(global_encoding), data


def _unpack_bytes(data: bytes) -> (bytes, bytes):
    length = struct.unpack("i", data[:4])[0]
    data = data[4:]
    return (
        struct.unpack("{}s".format(length), data[:length])[0],
        data[length:],
    )


def _unpack_tuple(data: bytes, t: Tuple[Any]) -> (Tuple[Any], bytes):
    unpacked_objs: List[Any] = []
    for obj in t:
        (result, data) = _unpack_helper(data, obj)
        unpacked_objs.append(result)
    return tuple(unpacked_objs), data


def _unpack_list(data: bytes, t: List[T]) -> (List[T], bytes):
    length = struct.unpack("i", data[:4])[0]
    data = data[4:]
    obj_shell = t[0]
    t.pop()
    for i in range(0, length):
        (result, data) = _unpack_helper(data, obj_shell)
        t.append(result)
    return t, data


def _unpack_helper(data: bytes, obj: T) -> (T, bytes):
    if type(obj) in unpack_dict:
        return unpack_dict[type(obj)](obj, data)
    else:
        for v in vars(obj):
            (result, data) = _unpack_helper(data, obj.__getattribute__(v))
            obj.__dict__[v] = result
    return obj, data


@check_encoding
def load(*objs: Any, encoding: str = "utf-8") -> bytes:
    """
    Converts *objs* into a byte string format in order.

    :param encoding: encoding for strings to be encoded in
    :param objs: Objects to be converted to byte string
    :return: Byte string of *objs*
    """
    global global_encoding
    global_encoding = encoding
    return b"".join(
        [
            pack_dict.get(
                type(obj),
                lambda o: b"".join([load(o.__getattribute__(v)) for v in vars(o)]),
            )(obj)
            for obj in objs
        ]
    )


@check_encoding
def unload(data: bytes, *objs: Any, encoding: str = "utf-8") -> Tuple[Any]:
    """
    Converts *data* into python objects, in the order and format of *objs*

    :param encoding: encoding that strings are encoded in
    :param data: data to be converted
    :param objs: order and objects that data conforms to
    :return: One object if *objs* contains one element, or tuple of objects if *objs* contains more than 1 element
    """
    global global_encoding
    global_encoding = encoding
    if len(objs) == 0:
        raise TypeError("unload() takes a variable number of objects")
    if len(objs) == 1:
        return _unpack_helper(data, objs[0])[0]
    else:
        unpacked_objs: List[Any] = []
        for obj in objs:
            if len(data) == 0:
                raise Exception("Data did not match objects.")
            (result, data) = _unpack_helper(data, obj)
            unpacked_objs.append(result)
        if len(data) != 0:
            raise Exception(f"Data did not match objects.")
        return tuple(unpacked_objs)



from .__version__ import (
    __title__,
    __description__,
    __url__,
    __version__,
    __author__,
    __author_email__,
    __license__,
)

from .api import (
    load,
    unload,
)
from .strict_type import (
    Char,
    UnsignedChar,
    SignedChar,
    Short,
    UnsignedShort,
    Int,
    UnsignedInt,
    Long,
    UnsignedLong,
    LongLong,
    UnsignedLongLong,
    Float,
    Double,
)



__title__ = "pyparcel"
__description__ = "The simple and secure way to convert objects to bytestrings."
__url__ = "https://najaco.github.io/pyparcel/"
__version__ = "1.0.0"
__author__ = "Nathan Cohen"
__author_email__ = "ncohen4299@gmail.com"
__license__ = "MIT License"



import struct
from abc import ABC, abstractmethod
from typing import Any, TypeVar

T = TypeVar("T")


class StrictType(ABC):
    """
    Abstract class for StrictType. Defines standard ``__pack__`` and ``__unpack__`` functions.
    :cvar FORMAT: Format specifier used when packing obj
    :cvar STANDARD_SIZE: Standard size of obj in bytes
    """

    @property
    @abstractmethod
    def FORMAT(self):
        pass

    @property
    @abstractmethod
    def STANDARD_SIZE(self):
        pass

    def __pack__(self) -> bytes:
        return struct.pack(self.FORMAT, self)

    def __unpack__(self, data: bytes) -> (T, bytes):
        return (
            struct.unpack(self.FORMAT, data[: self.STANDARD_SIZE])[0],
            data[self.STANDARD_SIZE :],
        )


class Char(str, StrictType):
    """
    Represents the strict type for a char.
    """

    FORMAT: str = "=c"
    STANDARD_SIZE: int = 1

    def __init__(self, c: str = "\0"):
        """
        Initializes
        :param c:
        """
        if len(c) != 1:
            raise Exception("Char may only take arguments of length 1")
        super(str, Char).__init__(c)

    def __pack__(self) -> bytes:
        return struct.pack(self.FORMAT, self.encode("ascii"))

    def __unpack__(self, data: bytes) -> (float, bytes):
        return (
            struct.unpack(self.FORMAT, data[: self.STANDARD_SIZE])[0].decode("ascii"),
            data[self.STANDARD_SIZE :],
        )


class UnsignedChar(int, StrictType):
    """
    Represents the strict type for an unsigned char.
    """

    FORMAT: str = "=B"
    STANDARD_SIZE: int = 1

    def __init__(self, c: int = 0):
        if not 0 <= c <= 255:
            raise Exception("UnsignedChar must be between 0 and 255")
        super(int, UnsignedChar).__init__(c)


class SignedChar(int, StrictType):
    """
    Represents the strict type for a signed char.
    """

    FORMAT: str = "=b"
    STANDARD_SIZE: int = 1

    def __init__(self, c: int = 0):
        if not -128 <= c <= 127:
            raise Exception("SignedChar must be between -128 and 127")
        super(int, SignedChar).__init__(c)


class Short(int, StrictType):
    """
    Represents the strict type for a short.
    """

    FORMAT: str = "=h"
    STANDARD_SIZE: int = 2

    def __init__(self, n: int = 0):
        if not -32768 <= n <= 32767:
            raise Exception("Short must be between -32768 and 32767")
        super(int, Short).__init__(n)


class UnsignedShort(int, StrictType):
    """
    Represents the strict type for an unsigned short.
    """

    FORMAT: str = "=H"
    STANDARD_SIZE: int = 2

    def __init__(self, n: int = 0):
        if not 0 <= n <= 65535:
            raise Exception("Short must be between 0 and 65535")
        super(int, UnsignedShort).__init__(n)


class Int(int, StrictType):
    """
    Represents the strict type for an integer.
    """

    FORMAT: str = "=i"
    STANDARD_SIZE: int = 4

    def __init__(self, n: int = 0):
        if not -2147483648 <= n <= 2147483647:
            raise Exception("Short must be between -2147483648 and 2147483647")
        super(int, Int).__init__(n)


class UnsignedInt(int, StrictType):
    """
    Represents the strict type for an unsigned integer.
    """

    FORMAT: str = "=I"
    STANDARD_SIZE: int = 4

    def __init__(self, n: int = 0):
        if not 0 <= n <= 4294967295:
            raise Exception("Unsigned Integer must be between 0 and 4294967295")
        super(int, UnsignedInt).__init__(n)


class Long(int, StrictType):
    """
    Represents the strict type for an long.
    """

    FORMAT: str = "=l"
    STANDARD_SIZE: int = 4

    def __init__(self, n: int = 0):
        if not -2147483648 <= n <= 2147483647:
            raise Exception("Long must be between -2147483648 and 2147483647")
        super(int, Long).__init__(n)


class UnsignedLong(int, StrictType):
    """
    Represents the strict type for an unsigned long.
    """

    FORMAT: str = "=L"
    STANDARD_SIZE: int = 4

    def __init__(self, n: int = 0):
        if not 0 <= n <= 4294967295:
            raise Exception("UnsignedLong must be between 0 and 4294967295")
        super(int, UnsignedLong).__init__(n)


class LongLong(int, StrictType):
    """
    Represents the strict type for a long long.
    """

    FORMAT: str = "=q"
    STANDARD_SIZE: int = 8

    def __init__(self, n: int = 0):
        if not -9223372036854775808 <= n <= 9223372036854775807:
            raise Exception("LongLong must be between 0 and 9223372036854775807")
        super(int, LongLong).__init__(n)


class UnsignedLongLong(int, StrictType):
    """
    Represents the strict type for an unsigned long long.
    """

    FORMAT: str = "=Q"
    STANDARD_SIZE: int = 8

    def __init__(self, n: int = 0):
        if not 0 <= n <= 18446744073709551615:
            raise Exception(
                "UnsignedLongLong must be between 0 and 18446744073709551615"
            )
        super(int, UnsignedLongLong).__init__(n)


class Float(float, StrictType):
    """
    Represents the strict type for a float.
    """

    FORMAT: str = "=f"
    STANDARD_SIZE: int = 4

    def __init__(self, n: int = 0):
        super(float, Float).__init__(n)


class Double(float, StrictType):
    """
    Represents the strict type for a double.
    """

    FORMAT: str = "=d"
    STANDARD_SIZE: int = 8

    def __init__(self, n: int = 0):
        super(float, Double).__init__(n)



import pyparcel


class Foo:
    def __init__(self, a: int = int(), b: float = float(), c: str = str()):
        self.a: int = a
        self.b: float = b
        self.c: str = c

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.a == other.a and self.b == other.b and self.c == other.c
        return False


foo: Foo = Foo(8, 5.7, "Hello World")  # Foo(int, float, str)
data: bytes = pyparcel.load(foo)  # b'\x08\x00\x00\x00ff\xb6@\x0b\x00\x00\x00Hello World'
# ...
baz: Foo = pyparcel.unload(data, Foo())  # foo == baz



from pyparcel import Long, UnsignedLongLong, Double


class Foo:
    def __init__(self, a: int, b: int, c: int, d: long):
        self.a = a
        self.b = Long(b)
        self.c = UnsignedLongLong(c)
        self.d = Double(d)



import setuptools
import pyparcel

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=pyparcel.__title__,
    version=pyparcel.__version__,
    author=pyparcel.__author__,
    author_email=pyparcel.__author_email__,
    description="Easy to use binary packer that doesn't require struct's specifications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/najaco/pyparcel",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
